# Model the fit results with a function
# Derek Fujimoto
# Nov 2019


from tkinter import *
from tkinter import ttk
from functools import partial

import logging,re,os,warnings
import numpy as np
import pandas as pd
import bdata as bd

from bfit.backend.ConstrainedFunction import ConstrainedFunction as CstrFnGenerator
from bfit.fitting.fit_bdata import fit_bdata
from bfit.gui.template_fit_popup import template_fit_popup

# ========================================================================== #
class popup_fit_constraints(template_fit_popup):
    """
        Popup window for modelling the fit results with a function
        
        bfit
        fittab
        logger
        
        output_par_text     text, detected parameter names
        output_text         dict, keys: p0,blo,bhi,res,err, value: tkk.Text objects
       
        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text
       
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}
    
    window_title = 'Fit data with contrained parameters'
    
    # ====================================================================== #
    def __init__(self,bfit,constr_text='',output_par_text='',output_text=''):
        
        super().__init__(bfit,constr_text,output_par_text,output_text)
        
        # Keyword parameters
        key_param_frame = ttk.Frame(self.left_frame,relief='sunken',pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = CstrFnGenerator.keyvars
        
        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len,keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk),d) for k,d in zip(keys,descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame,text=s,justify=LEFT)
        
        # fit parameter names 
        fit_param_frame = ttk.Frame(self.left_frame,relief='sunken',pad=5)
        s = 'Reserved function parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(),
                                        self.fittab.n_component.get())
        
        s += '\n'.join([k for k in sorted(self.parnames)]) 
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame,text=s,justify=LEFT)

        # module names 
        module_frame = ttk.Frame(self.left_frame,relief='sunken',pad=5)
        s = 'Reserved module names:\n\n'
        
        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len,keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk),d) for k,d in zip(keys,descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame,text=s,justify=LEFT)
        
        # Text entry
        self.entry_label['text'] = 'Enter one constraint equation per line.'+\
                                 '\nNon-reserved words are shared variables.'+\
                                 '\nEx: "1_T1 = a*np.exp(b*BIAS**0.5)+c"'+\
                                 '\nNote: Shared and fixed flags from main window ignored.'
                
        # gridding
        key_param_label.grid(column=0,row=0)
        fit_param_label.grid(column=0,row=0)
        modules_label.grid(column=0,row=0)
        
        key_param_frame.grid(column=0,row=0,rowspan=1,sticky=(E,W),padx=1,pady=1)
        module_frame.grid(column=0,row=1,sticky=(E,W),padx=1,pady=1,rowspan=2)
        fit_param_frame.grid(column=0,row=3,sticky=(E,W,N,S),padx=1,pady=1)
        
    # ====================================================================== #
    def _do_fit(self,text):
        """
            Set up the fit functions and do the fit. Then map the outputs to the
            proper displays. 
        """
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # check that the defined variables all match function inputs
        for d in defined: 
            if d not in self.parnames:
                errmsg = 'Definition for "%s" invalid. ' % d+\
                         'Must only define function inputs. '
                messagebox.showerror("Error",errmsg)
                raise RuntimeError(errmsg)
        
        # make shared parameters for the rest of the parameters
        allpar = self.new_par['name'].tolist()
        alldef = defined[:]     # all parameter names in order
        sharelist = [True]*len(allpar)
        
        for n in sorted(self.parnames):
            if n not in defined:
                eqn.append(n)
                alldef.append(n)
                allpar.append(n)
                sharelist.append(False)
                        
        # replace 1_T1 with lambda1
        for i,_ in enumerate(allpar):
            if '1_T1' in allpar[i]:
                allpar[i] = allpar[i].replace('1_T1','lambda1')
        
        for i,_ in enumerate(eqn):
            while '1_T1' in eqn[i]:
                eqn[i] = eqn[i].replace('1_T1','lambda1')
                
        # make constrained functions
        cgen= CstrFnGenerator(alldef, eqn, allpar, self.parnames)
        
        # get the functions and initial parameters
        fit_files = self.bfit.fit_files
        fetch_files = self.bfit.fetch_files
        fitfns = []
        par = []
        rebin = []
        omit = []
        fnptrs = []
        constr_fns = []
        
        keylist = sorted(fit_files.fit_lines.keys())
        for k in keylist:
            line = fetch_files.data_lines[k]
            data = line.bdfit
            
            # get pulse length
            pulse_len = -1
            try:
                pulse_len = data.bd.get_pulse_s()
            except KeyError:
                pass
            
            # get function
            fn = fit_files.fitter.get_fn(fn_name=fit_files.fit_function_title.get(),
                                         ncomp=fit_files.n_component.get(),
                                         pulse_len=pulse_len,
                                         lifetime=bd.life[fit_files.probe_label['text']])
            
            genf, genc = cgen(data=data, fn=fn)
            fitfns.append(genf)
            fnptrs.append(fn)
            constr_fns.append(genc)
            
            # get initial parameters
            par.append(data.fitpar)
            
            # get rebin
            rebin.append(data.rebin.get())
            
            # get bin omission
            omit.append(data.omit.get())
        
        # clean up omit strings
        for i,om in enumerate(omit):
            if om == fetch_files.bin_remove_starter_line:
                omit[i] = ''
        
        # set up p0, bounds
        p0 = self.new_par['p0'].values
        blo = self.new_par['blo'].values
        bhi = self.new_par['bhi'].values
        
        p0 = [[p]*len(keylist) for p in p0]
        blo = [[p]*len(keylist) for p in blo]
        bhi = [[p]*len(keylist) for p in bhi]
                
        for n in sorted(self.parnames):
            if n not in defined:
                p0.append( [p['p0' ][n] for p in par])
                blo.append([p['blo'][n] for p in par])
                bhi.append([p['bhi'][n] for p in par])
        
        p0 = np.array(p0).T
        blo = np.array(blo).T
        bhi = np.array(bhi).T
        
        # set up fitter inputs
        npar = len(sharelist)
        bounds = [[l,h] for l,h in zip(blo,bhi)]
        data = [self.bfit.data[k] for k in keylist]
        kwargs = {'p0':p0, 'bounds':bounds}
        
        # get minimizer
        if 'trf'   in self.bfit.fit_files.fitter.__name__:  minimizer = 'trf'
        if 'minos' in self.bfit.fit_files.fitter.__name__:  minimizer = 'migradminos'
        if 'hesse' in self.bfit.fit_files.fitter.__name__:  minimizer = 'migradhesse'
            
        
        # do the fit and kill fitting window
        par, std_l, std_u, cov, chi, gchi = fit_bdata(
                                                data=data,
                                                fn=fitfns,
                                                shared=sharelist,
                                                asym_mode='c',
                                                rebin=rebin,
                                                omit=omit,
                                                xlims=None,
                                                hist_select=self.bfit.hist_select,
                                                minimizer=minimizer,
                                                **kwargs)
        
        # calculate original parameter equivalents
        for i,k in enumerate(keylist):
            data = fetch_files.data_lines[k].bdfit
            
            # calculate
            old_par = [cfn(*par[i]) for cfn in constr_fns[i]]
            old_std_l = [st if adef not in defined else np.nan for st, adef in zip(std_l[i], alldef)] 
            old_std_l += [np.nan]*(len(alldef)-len(old_std_l))
            
            old_std_u = [st if adef not in defined else np.nan for st, adef in zip(std_u[i], alldef)]
            old_std_u += [np.nan]*(len(alldef)-len(old_std_u))
            
            old_chi = chi[i]
            
            # sort by original parameter name order
            old_std_l = [old_std_l[alldef.index(n)] for n in cgen.oldpar]
            old_std_u = [old_std_u[alldef.index(n)] for n in cgen.oldpar]
            
            # set to fitdata containers
            # [(parname),(par),(err-),(err+),chi,fnpointer]
            data.set_fitresult([cgen.oldpar, old_par, old_std_l, old_std_u, old_chi, fnptrs[i]])
            
        # display in fit_files tab
        for key in fit_files.fit_lines:
            fit_files.fit_lines[key].show_fit_result()
        
        # show global chi
        fit_files.gchi_label['text'] = str(np.around(gchi, 2))

        # do end-of-fit stuff
        fit_files.do_end_of_fit()
        
        return (par[0,:], std_l[0,:], std_u[0,:])



# Function constrained by user-defined function
# Derek Fujimoto
# August 2019

import numpy as np

# =========================================================================== # 
class ConstrainedFunction(object):
    """
        p1
        p2
        constraints
        defined
        equation
    """
    
    # keywords used to identify variables
    keyvars = { 'B0'    : 'B0 Field (T)', 
                'BIAS'  : 'Platform Bias (kV)',
                'CLFT'  : 'Cryo Lift Read (mm)',
                'DUR'   : 'Run Duration (s)',
                'ENRG'  : 'Impl. Energy (keV)',
                'LAS'   : 'Laser Power',
                'NBMR'  : 'NBM Rate (count/s)',
                'RATE'  : 'Sample Rate (count/s)',
                'RF'    : 'RF Level DAC',
                'RUN'   : 'Run Number',
                'TEMP'  : 'Temperature (K)',
                'TIME'  : 'Start Time',
                'YEAR'  : 'Year',
              }    
                       
    # ======================================================================= # 
    def __init__(self, defined, equation, newpar, oldpar):
        """
            defined:        parameters which the equations define (equation LHS)
                            in old parameter order
            equation:       list of strings corresponding to equation RHS in old 
                            parameter order
            newpar:         list of strings corresponding to new function 
                            parameters in order
            oldpar:         list of strings corresponding to old function 
                            parameters in order
        """
        self.header = 'lambda %s : ' % (','.join(newpar))
        self.oldpar = oldpar
        
        # sort equations and defined by old par
        self.equation = [equation[defined.index(p)] for p in oldpar]
        
    # ======================================================================= # 
    def __call__(self, data, fn):
        """ 
            Identify variable names, make constraining function
            
            data: bfitdata object to generate the constrained function 
            fn: function handle
        """
        
        # get variables in decreasing order of length (no mistakes in replace)
        varlist = np.array(list(self.keyvars.keys()))
        varlist = varlist[np.argsort(list(map(len,varlist))[::-1])]
    
        eqn = []
        for c in self.equation:
                
            # find constant names in the string, replace with constant
            for var in varlist:
                if var in c:
                    value = self._get_value(data, var)
                    c = c.replace(var, str(value))
            eqn.append(c)
        
        # get constraint functions, sorted 
        constr_fns = [eval(self.header+e)for e in eqn]    
        
        # define the new fitting function
        def new_fn(x,*newparam):
            oldparam = [c(*newparam) for c in constr_fns]
            return fn(x, *oldparam)
            
        return (new_fn, constr_fns)
            
    # ======================================================================= # 
    def _get_value(self, data, name):
        """
            Tranlate typed constant to numerical value
        """
        
        if   name == 'B0'   :   return data.field
        elif name =='BIAS'  :   return data.bias
        elif name =='CLFT'  :   return data.bd.camp.clift_read.mean
        elif name =='DUR'   :   return data.bd.duration
        elif name =='ENRG'  :   return data.bd.beam_kev()
        elif name =='LAS'   :   return data.bd.epics.las_pwr.mean
        elif name =='NBMR'  :   
            return np.sum([data.hist['NBM'+h].data \
                           for h in ('F+','F-','B-','B+')])/data.duration
        elif name =='RATE'  :   
            hist = ('F+','F-','B-','B+') if data.area == 'BNMR' \
                                         else ('L+','L-','R-','R+')    
            return np.sum([data.hist[h].data for h in hist])/data.duration
        elif name =='RF'    :   return data.bd.camp.rf_dac.mean
        elif name =='RUN'   :   return data.run
        elif name =='TEMP'  :   return data.temperature.mean
        elif name =='TIME'  :   return data.bd.start_time
        elif name =='YEAR'  :   return data.year



# Calculate needed current from desired magnetic field. 
# Derek Fujimoto
# Feb 2019

from tkinter import *
from tkinter import ttk
import pandas as pd
import numpy as np
import logging
from bfit import logger_name
import bfit

class popup_show_param(object):
    """
        Display pandas dataframe object as pretty table in popup window.
    """
    
    rounding = 8
    
    # ======================================================================= #
    def __init__(self,df):
        """
            Draw window for show parameter popup
            
            df: pandas dataframe with run information
        """
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # root 
        root = Toplevel()
        root.title("Fitted Parameters")
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)
        
        # icon
        try:
            img = PhotoImage(file=bfit.icon_path)
            root.tk.call('wm', 'iconphoto', root._w, img)
        except Exception as err:
            print(err)
        
        # main frame
        mainframe = ttk.Frame(root,pad=5)
        mainframe.grid(column=0,row=0,sticky=(N,W))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)
        
        # round values
        df.reset_index(inplace=True)
        df = df.round(decimals=self.rounding)
        
        # make strings with errors
        for c in df.columns:
            if 'Error' in c:    continue
            
            if 'Error+ '+c in df.columns and 'Error- '+c in df.columns:
                df[c] = df[c].map(str) + ' [+' + df['Error+ '+c].map(str) + \
                                          ' -' + df['Error- '+c].map(str)+']'
            elif 'Error+ '+c in df.columns:
                df[c] = df[c].map(str) + ' +' + df['Error+ '+c].map(str)
            elif 'Error- '+c in df.columns:
                df[c] = df[c].map(str) + ' -' + df['Error- '+c].map(str)
            elif 'Error '+c in df.columns:
                df[c] = df[c].map(str) + ' +/- ' + df['Error '+c].map(str)
                                
        df.drop(labels=df.columns[['Error' in c for c in df.columns]],
                axis='columns',inplace=True)
        
        # make list of outputs
        header = df.columns.values
        output_list = df.values
        output_list = np.vstack((header,output_list))
        
        # get len of each column
        maxlen = [max(map(len,map(str,line))) for line in output_list.T]
        splitter = ['-'*n for n in maxlen]
        output_list = np.vstack((output_list[0],splitter,output_list[1:]))
        
        # get nicely formatted string to output
        str_fmt = ' '.join(['%-'+'%ds' % n for n in maxlen])
                    
        nice_output = '\n'.join([str_fmt % tuple(line) for line in output_list])
                
        # Text box and other objects
        self.text = Text(mainframe,width=150,height=30,state='normal',wrap="none")
        self.text.delete('1.0',END)
        self.text.insert('1.0',nice_output)
        
        vsb = ttk.Scrollbar(mainframe, orient="vertical")
        hsb = ttk.Scrollbar(mainframe, orient="horizontal")
        
        self.text.configure(yscrollcommand=vsb.set)
        vsb.configure(command=self.text.yview)
        self.text.configure(xscrollcommand=hsb.set)
        hsb.configure(command=self.text.xview)
        
        # Gridding
        self.text.grid(column=0,row=0,sticky=(N,S,E,W))
        vsb.grid(row=0, column=1, sticky="nse")
        hsb.grid(row=1, column=0, sticky="wse")
        
        mainframe.grid_columnconfigure(0,weight=1)
        mainframe.grid_rowconfigure(0,weight=1)
        
        # runloop
        self.root = root
        self.logger.debug('Initialization success. Starting mainloop.')
        root.mainloop()
    



# Calculate needed Vpp from desired H1 magnetic field. 
# Derek Fujimoto
# December 2017

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import numpy as np
import webbrowser
import logging
import bfit

# =========================================================================== #
class calculator_nmr_B1(object):
    
    # ======================================================================= #
    def __init__(self):
        """Draw window for calculating needed Vpp from desired H1 magnetic field."""
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # root 
        root = Toplevel()
        root.title("Gerald's Caculator")
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)
        
        # icon
        try:
            img = PhotoImage(file=bfit.icon_path)
            root.tk.call('wm', 'iconphoto', root._w, img)
        except Exception as err:
            print(err)
        
        # key bindings
        root.bind('<Return>',self.calculate)             
        root.bind('<KP_Enter>',self.calculate)
        
        # variables
        self.field = StringVar()
        self.field.set("")
        self.volt = StringVar()
        self.volt.set("")
        self.nu = StringVar()
        
        self.nu.set('41.27') # MHz
        
        # main frame
        mainframe = ttk.Frame(root,pad=5)
        mainframe.grid(column=0,row=0,sticky=(N,W))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)
        
        # Entry and other objects
        title_line = ttk.Label(mainframe,   
                text='BNMR Magnetic Oscillating Field -- '+\
                     'Antenna Voltage (peak-peak)', 
                     justify=CENTER)
        self.entry_field = ttk.Entry(mainframe,textvariable=self.field,width=10,
                justify=RIGHT)
        gauss = ttk.Label(mainframe,text='Gauss')
        equals = ttk.Label(mainframe,text='=')
        self.entry_voltage = ttk.Entry(mainframe,textvariable=self.volt,
                width=10,justify=RIGHT)
        voltage = ttk.Label(mainframe,text='millivolts')
        explanation = ttk.Label(mainframe,text='Press Enter to convert',
                justify=CENTER)
        
        # frequency input
        freq_frame = ttk.Frame(root,pad=5)
        
        freq1 = ttk.Label(freq_frame,text='Freq:')
        self.entry_freq = ttk.Entry(freq_frame,textvariable=self.nu,width=10,
                                    justify=RIGHT)
        freq2 = ttk.Label(freq_frame,text='MHz')
        
        freq1.grid(column=0,row=0,padx=5)
        self.entry_freq.grid(column=1,row=0,padx=5)
        freq2.grid(column=2,row=0,padx=5)
        
        # caveat
        warning = ttk.Label(root,
                            text='WARNING: Calculation only valid for runs \n'+\
                                 'with silver antenna (before Sept 2019)', 
                            justify=CENTER)
        
        # Gridding
        title_line.grid(        column=0,row=0,padx=5,pady=5,columnspan=5)
        freq_frame.grid(        column=0,row=1,padx=5,pady=5,columnspan=5)
        self.entry_field.grid(  column=0,row=2,padx=5,pady=5,sticky=E)
        gauss.grid(             column=1,row=2,padx=5,pady=5,sticky=E)
        equals.grid(            column=2,row=2,padx=20,pady=5,sticky=(E,W))
        self.entry_voltage.grid(column=3,row=2,padx=5,pady=5,sticky=W)
        voltage.grid(           column=4,row=2,padx=5,pady=5,sticky=W)
        explanation.grid(       column=0,row=3,padx=5,pady=5,columnspan=5)
        warning.grid(           column=0,row=2,padx=5,pady=5,columnspan=5)
        
        # runloop
        self.root = root
        self.logger.debug('Initialization success. Starting mainloop.')
        root.mainloop()
        
    # ======================================================================= #
    def calculate(self,*args):
        
        # check focus
        focus_id = str(self.root.focus_get())
        
        # get freqeuency
        nu = float(self.nu.get())
        
        # convert field to voltage
        if focus_id == str(self.entry_field):        
            try:
                field = float(self.field.get())
                value = field/0.0396*nu
                self.volt.set("%9.6f" % np.around(value,6))
                
                self.logger.info('Field of %g G converted to voltage of %g mV',
                                 field,value)
            except ValueError:
                self.logger.exception('Bad input.')
                self.volt.set('Error')
        
        # convert voltage to field
        elif focus_id == str(self.entry_voltage):        
            try:
                voltage = float(self.volt.get())
                value = voltage*0.0396/nu
                self.field.set("%9.6f" % np.around(value,6))
                
                self.logger.info('Voltage of %g mV converted to field of %g G',
                                 voltage,value)
            except ValueError:
                self.logger.exception('Bad input.')
                self.field.set('Error')
            





# Set function paramters from gui window
# Derek Fujimoto
# April 2019

from tkinter import *
from tkinter import ttk
from bfit import logger_name
from bfit.backend.FunctionPlacer import FunctionPlacer
from bfit.fitting.decay_31mg import fa_31Mg
import bfit.fitting.functions as fns

import matplotlib.pyplot as plt
import logging
import bdata as bd
import numpy as np

# ========================================================================== #
class popup_param(object):
    """
        Popup window for graphically finding input parameters. 
        
        data:           bdata object 
        fig:            maplotlib figure object
        fitter:         fit_tab.fitter obje (defined in default_routines.py)
        first:          if True, first time through fixing parameters
        fname:          name of the function 
        logger:         logging variable
        mode:           1 or 2 to switch between run modes
        n_components:   number of components in the fit function
        parnames:       tuple of names of the parameters
        p0:             dictionary of StringVar objects to link parameters
        selection:      StringVar, track run selection
        win:            TopLevel window
        xy:             (x,asym,dasym) tuple
    """

    # parameter mapping
    parmap = {  '1_T1':'lam',
                '1_T1b':'lamb',
                'amp':'amp',
                'beta':'beta',
                'fraction_b':'fraction_b',
                'baseline':'base',
                'peak':'peak',
                'fwhmA':'fwhm',
                'fwhmB':'fwhm',
                'height':'amp',
                'heightA':'amp',
                'heightB':'amp',
                'sigma':'fwhm',
                'mean':'peak',
                'amp0':'amp0',
                'amp1':'amp1',
                'amp2':'amp2',
                'amp3':'amp3',
                'nu_0':'nu_0',
                'nu_q':'nu_q',
                'efgAsym':'eta',
                'efgPhi':'phi',
                'efgTheta':'theta',
                'fwhm':'fwhm',
             }

    # ====================================================================== #
    def __init__(self,bfit,id=''):
        self.bfit = bfit
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing gui param popup')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Find P0')
        frame = ttk.Frame(self.win,relief='sunken',pad=5)
        
        # icon 
        self.bfit.set_icon(self.win)
        
        # Labels
        ttk.Label(frame,text="Select Run").grid(column=0,row=0,sticky=E)
        
        # box for run select
        self.selection = StringVar()
        select_box = ttk.Combobox(frame,textvariable=self.selection,
                                  state='readonly')
        select_box.bind('<<ComboboxSelected>>',self.setup)
        
        # get run list
        runlist = list(self.bfit.fit_files.fit_lines.keys())
        runlist.sort()
        # ~ runlist = ['('+r.split('.')[0]+') '+r.split('.')[1] for r in runlist]
        select_box['values'] = runlist
        
        # gridding
        frame.grid(column=0,row=1,sticky=(N,W,E,S))
        select_box.grid(column=0,row=1,sticky=E)
        
        # start looking automatically
        if id:
            self.selection.set(id)
            self.setup()
            self.win.withdraw()
                
    # ====================================================================== #
    def setup(self,*args):
        """Get parameters for placing function and start the run squence"""
        
        # get run selection 
        run_id = self.selection.get()
        self.logger.info('Running P0 GUI finder on run %s',run_id)
        
        # get data
        self.data = self.bfit.data[run_id]
        mode = self.data.mode
        
        # mode switching
        if mode in ('20','2h'):         self.mode = 2
        elif mode in ('1f','2e','1w'):  self.mode = 1
        else:
            self.logger.warning('P0 Finder not configured for run mode %s',mode)
            print('P0 Finder not configured for run mode %s'%mode)
        
        # make new window 
        self.fig = plt.figure()
        ax = self.fig.add_subplot(111)
        
        # draw data
        omit = self.data.omit.get()
        if omit == self.bfit.fetch_files.bin_remove_starter_line:
            omit = ''
        self.xy = self.data.asym(self.bfit.get_asym_mode(self.bfit.fit_files),
                                 rebin=self.data.rebin.get(),omit=omit)
        ax.errorbar(*self.xy,fmt='.',color='k',ecolor='k')
        
        # plot elements - don't do tight_layout here - blocks matplotlib signals
        ax.set_ylabel('Asymmetry')
        if self.mode == 2:     ax.set_xlabel('Time (s)')
        elif self.mode == 1:   ax.set_xlabel('Frequency (MHz)')
        self.fig.show()
        
        # get parameters list and run finder 
        fit_tab = self.bfit.fit_files
        self.fitter = fit_tab.fitter
        self.n_components = fit_tab.n_component.get()
        self.fname = fit_tab.fit_function_title.get()
        self.parnames = self.fitter.gen_param_names(fn_name=self.fname,
                                          ncomp=self.n_components)
        parentry = fit_tab.fit_lines[run_id].parentry
        
        # make initial paramter list
        self.p0 = {k:parentry[k]['p0'][0] for k in parentry.keys()}
        
        self.run()
        
    # ====================================================================== #
    def run(self):
        """
            Run the function placer
            
            comp = component number to run
        """
        
        try:
            del self.fplace
        except AttributeError:
            pass
        else:
            self.fig.axes[0].cla()
            self.fig.axes[0].errorbar(*self.xy,fmt='.')
        
        # ensure matplotlib signals work. Not sure why this is needed.
        self.fig.tight_layout()
        
        # get paramters, translating the names
        p0 = []
        if self.n_components > 1:
            for i in range(self.n_components):
                p0_element = {}
                for k in self.p0.keys():
                    if 'base' in k:
                        p0_element[self.parmap[k]] = self.p0[k] 
                    elif str(i) in k:
                        p0_element[self.parmap['_'.join(k.split('_')[:-1])]] = self.p0[k] 
                p0.append(p0_element)
        else:
            p0.append({self.parmap[k]:self.p0[k] for k in self.p0.keys()})
        
        # get fitting function 
        if self.fname == 'Lorentzian':
            fn = fns.lorentzian     # freq,peak,fwhm,amp
        elif self.fname == 'BiLorentzian':
            fn = lambda freq,peak,fwhm,amp : fns.bilorentzian(freq,peak,fwhm,amp/3,fwhm/3,amp*2/3)
        elif self.fname == 'Gaussian':
            fn = lambda freq,peak,fwhm,amp : fns.gaussian(freq,peak,fwhm,amp)
        elif self.fname == 'QuadLorentz':
            fn = lambda freq, nu_0, nu_q, eta, theta, phi, \
                        amp0, amp1, amp2, amp3, fwhm: \
                        fns.quadlorentzian(freq, nu_0, nu_q, eta, theta, phi, \
                        amp0, amp1, amp2, amp3, \
                        fwhm, fwhm, fwhm, fwhm,
                        I = self.fitter.spin[self.fitter.probe_species])
                
        elif self.fname in ('Exp', 'Str Exp'):
            
            # get function
            pulse = self.data.get_pulse_s()
            lifetime = bd.life[self.bfit.probe_species.get()]
        
            if self.fname == 'Exp':
                f1 = fns.pulsed_exp(lifetime=lifetime,pulse_len=pulse)
                
                if self.bfit.probe_species.get() == 'Mg31':
                    fn = lambda x,lam,amp : fa_31Mg(x,pulse)*f1(x,lam,amp)
                else:
                    fn = lambda x,lam,amp : f1(x,lam,amp)
            
            elif self.fname == 'Str Exp':
                f1 = fns.pulsed_strexp(lifetime=lifetime,pulse_len=pulse)
                
                if self.bfit.probe_species.get() == 'Mg31':
                    fn = lambda x,lam,amp,beta : fa_31Mg(x,pulse)*f1(x,lam,beta,amp)
                else:
                    fn = lambda x,lam,amp,beta : f1(x,lam,beta,amp)
        else:
            self.cancel()
            errormsg = 'Function "%s" not implemented in P0 Finder' % self.fname
            self.logger.warning(errormsg)
            messagebox.showerror("Error",errormsg)
            raise RuntimeError(errormsg)
        
        self.fig.canvas.mpl_connect('close_event',self.cancel)
        
        self.fplace = FunctionPlacer(fig=self.fig,
                                     data=self.data,
                                     fn_single=fn,
                                     ncomp=self.n_components,
                                     p0=p0,
                                     fnname=self.fname,
                                     asym_mode=self.bfit.get_asym_mode(self.bfit.fit_files),
                                     endfn=self.endfn,
                                     spin=self.fitter.spin[self.fitter.probe_species])
        
    # ====================================================================== #        
    def endfn(self,p0,base):
        """Set output fields"""
        
        # single component
        if len(p0) == 1:
            p0 = p0[0]
            for k in self.p0.keys():
                if 'base' not in k:
                    self.p0[k].set(p0[self.parmap[k]])
        # multi component
        else:
            for k in self.p0.keys():
                if 'base' not in k:
                    s = k.split('_')
                    key = '_'.join(s[:-1])
                    i = s[-1]
                    self.p0[k].set(p0[int(i)][self.parmap[key]])
        
        # baseline
        if 'baseline' in self.p0.keys():
            self.p0['baseline'].set(base)
        
    # ====================================================================== #
    def cancel(self,*args):
        if hasattr(self,'fplace'):  del self.fplace
        if hasattr(self,'fig'):     
            plt.close(self.fig.number)
            del self.fig
        
        self.win.destroy()






# Wrapper for easy usage of the iminuit Minuit object
# Derek Fujimoto
# Nov 2020

import inspect
import numpy as np
from iminuit import Minuit
from bfit.fitting.leastsquares import LeastSquares

class minuit(Minuit):
    """
        Chi squared minimization for function of best fit using the minuit code
    """
    
    # ====================================================================== #
    def __init__(self, fn, x, y, dy=None, dx=None, dy_low=None, dx_low=None, 
                 fn_prime=None, fn_prime_dx=1e-6, name=None, start=None, 
                 error=None, limit=None, fix=None, print_level=1, **kwargs):
        """
            fn: function handle. f(x, a, b, c, ...)
            x:              x data
            y:              y data
            dy:             error in y
            dx:             error in x
            dy_low:         Optional, if error in y is asymmetric. If not none, dy is upper error
            dx_low:         Optional, if error in y is asymmetric. If not none, dx is upper error
            fn_prime:       Optional, function handle for the first derivative of fn. f'(x, a, b, c, ...)
            fn_prime_dx:    Spacing in x to calculate the derivative for default calculation
            name:           Optional sequence of strings. If set, use this for 
            start:          Optional sequence of numbers. Required if the 
                                function takes an array as input or if it has 
                                the form f(x, *pars), and name is not defined. 
                                Default: 1, broadcasted to all inputs
            error           Optional sequence of numbers. Initial step sizes. 
            limit           Optional sequence of limits that restrict the range 
                                format: [[low, high], [low, high], ...]
                                in which a parameter is varied by minuit. 
                                with None, +/- inf used to disable limit
            fix             Optional sequence of booleans. Default: False
            print_level Set the print_level
                            0 is quiet. 
                            1 print out at the end of MIGRAD/HESSE/MINOS. 
                            2 prints debug messages.
                
            kwargs:         passed to Minuit.from_array_func.
                
                To set for parameter "a" one can also assign the following 
                keywords instead of the array inputs
                    
                    a = initial_value
                    error_a = start_error
                    limit_a = (low, high)
                    fix_a = True
        """
        
        # construct least squares object
        ls = LeastSquares(fn = fn, 
                        x = x, 
                        y = y, 
                        dy = dy, 
                        dx = dx, 
                        dy_low = dy_low, 
                        dx_low = dx_low, 
                        fn_prime = fn_prime, 
                        fn_prime_dx = fn_prime_dx)

        # get number of data points
        self.npts = len(x)

        # detect function names
        if name is None:
            
            # get names from code
            name = inspect.getfullargspec(fn).args
            
            # remove self
            if 'self' in name: name.remove('self')
            
            # remove data input
            name = name[1:]
            
            # check for starting parameters
            if start is not None:
                
                # if they don't match assume array input to function
                try:
                    if len(start) != len(name):
                        name = ['x%d' for d in range(len(start))]
                
                # start is a scalar to be broadcasted
                except TypeError:
                    pass
                
            # check for bad input
            else:
                for n in name:
                    if '*' in n:
                        raise RuntimeError("If array input must define name or start")
                    
        # set errordef for least squares minimization
        kwargs['errordef'] = 1
        
        # set starting values, limits, fixed
        is_start = start is not None
        is_error = error is not None
        is_limit = limit is not None
        is_fix   = fix   is not None
        
        keys = kwargs.keys()
        
        # check limit depth 
        if is_start:    broadcast_start = get_depth(start) < 1
        if is_error:    broadcast_error = get_depth(error) < 1
        if is_limit:    broadcast_limit = get_depth(limit) < 2
        if is_fix:      broadcast_fix   = get_depth(fix)   < 1
        
        # iterate parameter names
        for n in name:
            
            # index of name 
            nidx = list(name).index(n)
            
            # start
            if n not in keys:
                
                if is_start:
                    
                    if broadcast_start: kwargs[n] = start
                    else:               kwargs[n] = start[nidx]
                        
                else:
                    kwargs[n] = 1
                    
            # error
            if is_error and 'error_'+n not in keys:
                
                if broadcast_error:     kwargs['error_'+n] = error
                else:                   kwargs['error_'+n] = error[nidx]
                    
            # limit
            if is_limit and 'limit_'+n not in keys:
                
                if broadcast_limit:     kwargs['limit_'+n] = limit
                else:                   kwargs['limit_'+n] = limit[nidx]    
            
            # fix
            if is_fix and 'fix_'+n not in keys:
                
                if broadcast_fix:       kwargs['fix_'+n] = fix
                else:                   kwargs['fix_'+n] = fix[nidx]
                
        # make minuit object
        super().__init__(ls, 
                         use_array_call=True, 
                         name = name, 
                         print_level = print_level,
                         **kwargs)
        
    # ====================================================================== #
    def chi2(self):
        nfixed = sum(self.fixed.values())
        dof = self.npts - self.narg + nfixed
        return self.fval/dof
        
def get_depth(lst, depth=0):
    try: 
        lst[0]
    except (IndexError, TypeError):
        return depth
    else:
        return get_depth(lst[0], depth+1)
        
        
        
        



import os
from .fitting.functions import lorentzian, bilorentzian, gaussian, quadlorentzian
from .fitting.functions import pulsed_exp, pulsed_strexp, pulsed_biexp
from .fitting.global_fitter import global_fitter
from .fitting.global_bdata_fitter import global_bdata_fitter
from .fitting.fit_bdata import fit_bdata
from .fitting.minuit import minuit

__all__ = ['gui','fitting','backend']
__version__ = '4.2.3'
__author__ = 'Derek Fujimoto'
logger_name = 'bfit'
icon_path = os.path.join(os.path.dirname(__file__),'images','icon.gif')

__all__.extend(("lorentzian", 
                "bilorentzian", 
                "gaussian", 
                "quadlorentzian",
                "pulsed_exp",
                "pulsed_strexp",
                "pulsed_biexp",
                "global_fitter",
                "global_bdata_fitter",
                "fit_bdata",
                "minuit",
                ))



# Fitter functions using curve_fit as the backend
# Derek Fujimoto
# Nov 2020

from bfit.fitting.fit_bdata import fit_bdata
from bfit.fitting.fitter import fitter as fit_base

class fitter(fit_base):
    
    __name__ = 'migrad (hesse)'
    
    def _do_fit(self, data, fn, omit=None, rebin=None, shared=None, hist_select='',
                xlims=None, asym_mode='c', fixed=None, parnames=None, **kwargs):
        """Inputs match fit_bdata"""
        
        return fit_bdata(data,
                         fn,
                         omit=omit,
                         rebin=rebin,
                         shared=shared,
                         hist_select=hist_select,
                         xlims=xlims,
                         asym_mode=asym_mode,
                         fixed=fixed,
                         minimizer='migrad',
                         name=parnames,
                         **kwargs)
            



from bfit.gui.bfit import bfit, logger_name
import logging, os, sys
from logging.handlers import RotatingFileHandler
import argparse, subprocess
from textwrap import dedent

# check if maxOS
if sys.platform == 'darwin':
    os.environ['OBJC_DISABLE_INITIALIZE_FORK_SAFETY'] = 'YES'

# now we can import multiprocessing
from multiprocessing import Process

# run
if __name__ == '__main__':

    # command line switches ---------------------------------------------------
    parser = argparse.ArgumentParser(description=dedent("""\
        Run BNMR data viewer and fitter for online application."""),
        formatter_class=argparse.RawTextHelpFormatter)
    
    # logging level
    parser.add_argument("-d", "--debug",
                        help='Run in debug mode',
                        dest='debug',
                        action='store_true',
                        default=False)

    # parse
    args = parser.parse_args()

    # Setup logging -----------------------------------------------------------
    logger = logging.getLogger(logger_name)

    # get log filename
    try:
        filename = os.path.join(os.environ['HOME'],'.bfit.log')
    except KeyError:
        filename = 'bfit.log'

    # make handler
    handler = RotatingFileHandler(filename,
                                  mode='a',
                                  maxBytes=100*1000, # 100 kB max
                                  backupCount=1)

    # get level and format for output string
    if args.debug:
        level = logging.DEBUG
        fmt = '%(asctime)s %(levelname)-8s %(module)s.%(funcName)s() [%(lineno)d] -- %(message)s'
    else:
        level = logging.INFO
        fmt = '%(asctime)s %(levelname)-8s %(module)s -- %(message)s'
    
    # set
    handler.setFormatter(logging.Formatter(fmt))
    handler.setLevel(level)

    logger.addHandler(handler)
    logger.setLevel(level)
    
    # testing
    testfn = None
    # ~ def testfn(self):
        # ~ self.fetch_files.run.set("40123, 40127")
        # ~ self.fetch_files.year.set(2012)
        # ~ self.fetch_files.get_data()
        # ~ self.fit_files.fit_function_title.set("QuadLorentz")
        # ~ self.fit_files.populate()
        # ~ self.notebook.select(2)
        # ~ self.fit_files.do_fit()
        # ~ import matplotlib.pyplot as plt
        # ~ plt.close('all')
        # ~ self.fit_files.do_add_param()
        
    # Check version ----------------------------------------------------------
    def check_version():
        try:
            vstr = subprocess.check_output([sys.executable,'-m','pip','search','bfit'])
        except subprocess.CalledProcessError:
            return
        else:            
            vstr = vstr.decode('utf-8')
            vlst = [l.strip() for l in vstr.split('\n') if l]
            
            # check if latest
            if 'latest' not in vlst[1]:
                print(vstr.strip())
    
    process_get_version = Process(target = check_version)
    process_get_version.start()
    
    # start bfit --------------------------------------------------------------
    bfit(testfn)
    
    # join
    process_get_version.join()
    



# Drawing style window
# Derek Fujimoto
# July 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import webbrowser
import logging

# ========================================================================== #
class popup_units(object):
    """
        Popup window for setting redraw period. 
        
        input:      dict:[StringVar,StringVar]
                    keyed by run mode, [conversion #, unit string]
    """

    # ====================================================================== #
    def __init__(self,parent):
        self.parent = parent    # fit_files pointer
        units = parent.units    # units dict from fit_files
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(parent.mainframe)
        self.win.title('Set Units')
        frame = ttk.Frame(self.win,relief='sunken',pad=5)
        topframe = ttk.Frame(frame,pad=5)

        # icon
        self.parent.set_icon(self.win)

        # Key bindings
        self.win.bind('<Return>',self.set)             
        self.win.bind('<KP_Enter>',self.set)

        # headers
        l_header = ttk.Label(topframe,text='Conversion',pad=5,justify=LEFT)
        l_header2 = ttk.Label(topframe,text='Unit',pad=5,justify=LEFT)

        # grid
        l_header.grid(column=1,row=0)
        l_header2.grid(column=2,row=0)

        # make entries
        self.input = {}    # [value StringVar, unit StringVar]
        r = 1
        for key in units.keys():
            
            l_val = ttk.Label(topframe,text=key,pad=5,justify=LEFT)
            self.input[key] = [StringVar(),StringVar()]
            
            for i in range(2):
                self.input[key][i].set(units[key][i])
                
            entry_val = ttk.Entry(topframe,textvariable=self.input[key][0],width=10,justify=RIGHT)        
            entry_uni = ttk.Entry(topframe,textvariable=self.input[key][1],width=5,justify=RIGHT)
            
            l_val.grid(column=0,row=r)
            entry_val.grid(column=1,row=r,padx=5)
            entry_uni.grid(column=2,row=r,padx=5)
            r += 1
            
        # make objects: buttons
        set_button = ttk.Button(frame,text='Set',command=self.set)
        close_button = ttk.Button(frame,text='Cancel',command=self.cancel)
        
        # grid
        topframe.grid(column=0,row=0,columnspan=2,pady=10)
        set_button.grid(column=0,row=r,pady=10)
        close_button.grid(column=1,row=r,pady=10)
            
        # grid frame
        frame.grid(column=0,row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def set(self,*args):
        """Set entered values"""     
           
        for key,value in self.input.items():
            self.parent.units[key] = [float(value[0].get()),value[1].get()]
        self.win.destroy()
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()



# Function for calculating new user-defined parameters
# Derek Fujimoto
# Nov 2020

import numpy as np
from bfit.backend.ConstrainedFunction import ConstrainedFunction

# =========================================================================== # 
class ParameterFunction(ConstrainedFunction):
    """
        name:           name of the parameter being defined
        inputs:         list of strings for parameter inputs to equation
        equation:       function handle, the equation defining the parameter
    """
    
    # ======================================================================= # 
    def __init__(self, name, equation, parnames, bfit):
        """
            bfit:           pointer to bfit object
            name:           name of parameter which we are defining (LHS)
            equation:       string corresponding to equation RHS
            parnames:       list of strings for fit parameter names
        """
        self.bfit = bfit
        
        # equations and names
        self.name = name
        
        # get variables in decreasing order of length (no mistakes in replace)
        varlist = np.array(list(self.keyvars.keys()))
        varlist = varlist[np.argsort(list(map(len,varlist))[::-1])]
    
        # replace 1_T1 with lambda
        while '1_T1' in equation:
            equation = equation.replace('1_T1', 'lambda1')
        
        self.parnames = list(parnames)
        for i,p in enumerate(self.parnames):
            while '1_T1' in p:
                p = p.replace('1_T1', 'lambda1')
            self.parnames[i] = p
    
        # make list of input names from meta data
        self.inputs = []
        
        for var in varlist:
            if var in equation:
                self.inputs.append(var)
                
        # make list of input names from parameter names
        for var in self.parnames:
            if var in equation:
                self.inputs.append(var)
                
        # make equation
        input_str = ', '.join(self.inputs)
        self.equation = eval('lambda %s : %s' % (input_str, equation))
        
        # add name to draw_components
        draw_comp = self.bfit.fit_files.draw_components
        if name in draw_comp:
            draw_comp.remove(name)
        draw_comp.append(name)
        
    # ======================================================================= # 
    def __call__(self):
        """ 
            Get data and calculate the parameter
        """
        
        inputs = {}
        runs = sorted(self.bfit.data.keys())
        for var in self.inputs:
            
            # get value for all data
            if var in self.keyvars:
                values = [self._get_value(self.bfit.data[r], var) for r in runs]
            elif var in self.parnames:
                
                var_par = var.replace('lambda1','1_T1')
                values = [self.bfit.data[r].fitpar['res'][var_par] for r in runs]
                    
            # set up inputs
            inputs[var] = np.array(values)
                            
        # calculate the parameter
        return self.equation(**inputs)
        



# Least squares class for minuit minimizer
# Derek Fujimoto
# Oct 2020

from scipy.misc import derivative 
import numpy as np


class LeastSquares:

    def __init__(self, fn, x, y, dy=None, dx=None, dy_low=None, dx_low=None, 
                 fn_prime=None, fn_prime_dx=1e-6):
        """
            fn: function handle. f(x, a, b, c, ...)
            x:              x data
            y:              y data
            dy:             error in y
            dx:             error in x
            dy_low:         used only if error in y is asymmetric. If not none, dy is upper error
            dx_low:         used only if error in y is asymmetric. If not none, dx is upper error
            fn_prime:       function handle for the first derivative of fn. f'(x, a, b, c, ...)
            fn_prime_dx:    spacing in x to calculate the derivative in the case of the default calculation
        """
        self.fn = fn
        self.x = x
        self.y = y
        self.n = len(x)
        
        # set derivative
        if fn_prime is None:
            self.fn_prime = lambda x, *pars : derivative(func=self.fn, x0=x, 
                                                         dx=fn_prime_dx, n=1, order=3, 
                                                         args=pars)
        else:
            self.fn_prime = fn_prime
        
        # set errors
        has_dy = False
        if dy is not None:
            has_dy = True
            self.dy = dy
        
        has_dx = False    
        if dx is not None:
            has_dx = True
            self.dx = dx
        
        # asymmetric errors
        has_dy_asym = False
        if dy_low is not None:
            
            if dy is None:
                has_dy = True
                self.dy = dy_low
            else:
                has_dy_asym = True
                self.dy_low = dy_low
        
        has_dx_asym = False
        if dx_low is not None:
            if dx is None:
                has_dx = True
                self.dx = dx_low
            else:
                has_dx_asym = True
                self.dx_low = dx_low
        
        # set least squares function
        if not any((has_dy, has_dx)):
            self.__call__ = self.ls_no_errors
        
        elif has_dx_asym and has_dy_asym:
            self.__call__ = self.ls_dxa_dya
        
        elif all((has_dx_asym, has_dy)) and not has_dy_asym:
            self.__call__ = self.ls_dxa_dy
        
        elif all((has_dy_asym, has_dx)) and not has_dx_asym:
            self.__call__ = self.ls_dx_dya
        
        elif has_dx_asym and not has_dy:
            self.__call__ = self.ls_dxa
        
        elif has_dy_asym and not has_dx:
            self.__call__ = self.ls_dya
        
        elif all((has_dy, has_dx)) and not any((has_dx_asym, has_dy_asym)):
            self.__call__ = self.ls_dxdy
        
        elif has_dy and not has_dx:
            self.__call__ = self.ls_dy

        elif has_dx and not has_dy:
            self.__call__ = self.ls_dx    
        
        else:
            raise RuntimeError("Missing error assignment case")
     
    def __call__(self, pars):
        return self.__call__(*pars)
        
    def ls_no_errors(self, *pars):
        return np.sum(np.square(self.y - self.fn(self.x, *pars)))
            
    def ls_dy(self, *pars):
        return np.sum(np.square((self.y -self.fn(self.x,*pars)) / self.dy))
    
    def ls_dx(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(self.dx*fprime)
        
        return np.sum(num/den)        

    def ls_dxdy(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(self.dx*fprime) + np.square(self.dy)
        return np.sum(num/den)   
             
    def ls_dya(self, *pars):
        
        # get errors on appropriate side of the function
        idx = self.y > self.fn(self.x, *pars)
        dy = np.array(self.dy)
        dy[idx] = self.dy_low[idx]
        
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(dy)
        return np.sum(num/den)        
  
    def ls_dxa(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        dx = 0.5*(self.dx+self.dx_low)
        
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(dx*fprime)
        return np.sum(num/den)        
       
    def ls_dx_dya(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        
        # get errors on appropriate side of the function
        idx = self.y > self.fn(self.x, *pars)
        dy = np.array(self.dy)
        dy[idx] = self.dy_low[idx]
        
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(self.dx*fprime) + np.square(dy)
        return np.sum(num/den)    
        
    def ls_dxa_dy(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        
        dx = 0.5*(self.dx+self.dx_low)
        
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(dx*fprime) + np.square(self.dy)
        return np.sum(num/den)    
        
    def ls_dxa_dya(self, *pars):
        fprime = self.fn_prime(self.x, *pars)
        dx = 0.5*(self.dx+self.dx_low)
        
        # get errors on appropriate side of the function
        idx = self.y > self.fn(self.x, *pars)
        dy = np.array(self.dy)
        dy[idx] = self.dy_low[idx]
        
        num = np.square(self.y - self.fn(self.x, *pars))
        den = np.square(dx*fprime) + np.square(dy)
        return np.sum(num/den)    
        



# Data object for holding bdata and related file settings for drawing and 
# fitting. 
# Derek Fujimoto
# Nov 2018

from tkinter import *
from bdata import bdata,bmerged
from bfit.gui.calculator_nqr_B0 import current2field
from bfit import logger_name

import numpy as np

import bfit
import logging
import textwrap

# =========================================================================== #
# =========================================================================== #
class fitdata(object):
    """
        Hold bdata and related file settings for drawing and fitting in fetch 
        files tab and fit files tab. 
        
        Data Fields:
            
            bfit:       pointer to top level parent object (bfit)
            bd:         bdata object for data and asymmetry (bdata)
            chi:        chisquared from fit (float)
            run:        run number (int)
            year:       run year (int)
            label:      label for drawing (StringVar)
            field:      magnetic field in T (float)
            field_std:  magnetic field standard deviation in T (float)
            bias:       platform bias in kV (float)
            bias_std:   platform bias in kV (float)
            
            id:         key for unique idenfication (str)    
            fitfnname:  function (str)
            fitfn:      function (function pointer)
            fitpar:     initial parameters {column:{parname:float}} and results
                        Columns are fit_files.fitinputtab.collist
            parnames:   parameter names in the order needed by the fit function
            
            drawarg:    drawing arguments for errorbars (dict)
            rebin:      rebin factor (IntVar)
            mode:       run mode (str)
            omit:       omit bins, 1f only (StringVar)
            check_state:(BooleanVar)    
    """
     
    # ======================================================================= #
    def __init__(self,parentbfit,bd):
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing run %d (%d).',bd.run,bd.year)
        
        # top level pointer
        self.bfit = parentbfit
        
        # bdata access
        self.bd = bd
        
        # input variables for tkinter
        self.rebin = IntVar()
        self.omit = StringVar()
        self.label = StringVar()
        self.check_state = BooleanVar()
        
        self.check_state.set(False)
        
        # fit parameters dictionary
        self.fitpar = {}
        
        # key for IDing file 
        self.id = self.bfit.get_run_key(data=bd)
        
        # initialize fitpar with fitinputtab.collist
        for k in ['p0','blo','bhi','res','dres+','dres-','chi','fixed','shared']:
            self.fitpar[k] = {}
        
        self.read()

    # ======================================================================= #
    def __getattr__(self,name):
        """Access bdata attributes in the case that fitdata doesn't have it."""
        try:
            return self.__dict__[name]
        except KeyError:
            return getattr(self.bd,name)

    # ======================================================================= #
    def asym(self,*args,**kwargs):  return self.bd.asym(*args,**kwargs)

    # ======================================================================= #
    def get_temperature(self,channel='A'):
        """
            Get the temperature of the run.
            Return (T, std T)
        """
        
        try:
            if channel == 'A':
                T = self.bd.camp['smpl_read_A'].mean
                dT = self.bd.camp['smpl_read_A'].std
            elif channel == 'B':
                T = self.bd.camp['smpl_read_B'].mean
                dT = self.bd.camp['smpl_read_B'].std
            elif channel == '(A+B)/2':
                Ta = self.bd.camp['smpl_read_A'].mean
                Tb = self.bd.camp['smpl_read_B'].mean
                dTa = self.bd.camp['smpl_read_A'].std
                dTb = self.bd.camp['smpl_read_B'].std
                
                T = (Ta+Tb)/2
                dT = ((dTa**2+dTb**2)**0.5)/2
            else:
                raise AttributeError("Missing required temperature channel.")
        
        except KeyError:
            T = np.nan
            dT = np.nan
        
        return (T,dT)
        
    # ======================================================================= #
    def read(self):
        """Read data file"""
        
        # bdata access
        if type(self.bd) is bdata:
            self.bd = bdata(self.run,self.year)
        elif type(self.bd) is bmerged:
            years = list(map(int,textwrap.wrap(str(self.year),4)))
            runs = list(map(int,textwrap.wrap(str(self.run),5)))
            self.bd = bmerged([bdata(r,y) for r,y in zip(runs,years)])
                
        # set temperature 
        try:
            self.temperature = temperature_class(*self.get_temperature(self.bfit.thermo_channel.get()))
        except AttributeError as err:
            self.logger.exception(err)
            try:
                self.temperature = self.bd.camp.oven_readC
            except AttributeError:
                self.logger.exception('Thermometer oven_readC not found')
                self.temperature = -1111
        
        # field
        try:
            if self.bd.area == 'BNMR':
                self.field = self.bd.camp.b_field.mean
                self.field_std = self.bd.camp.b_field.std
            else:
                self.field = current2field(self.bd.epics.hh_current.mean)*1e-4
                self.field_std = current2field(self.bd.epics.hh_current.std)*1e-4
        except AttributeError:
            self.logger.exception('Field not found')
            self.field = np.nan
            self.field_std = np.nan
            
        # bias
        try:
            if self.bd.area == 'BNMR': 
                self.bias = self.bd.epics.nmr_bias.mean
                self.bias_std = self.bd.epics.nmr_bias.std
            else:
                self.bias = self.bd.epics.nqr_bias.mean/1000.
                self.bias_std = self.bd.epics.nqr_bias.std/1000.
        except AttributeError:
            self.logger.exception('Bias not found')
            self.bias = np.nan

    # ======================================================================= #
    def set_fitpar(self,values):
        """Set fitting initial parameters
        values: output of routine gen_init_par: 
                {par_name:(par,lobnd,hibnd)}
        """
    
        self.parnames = list(values.keys())
    
        for v in values.keys():
            self.fitpar['p0'][v] = values[v][0]
            self.fitpar['blo'][v] = values[v][1]
            self.fitpar['bhi'][v] = values[v][2]
            self.fitpar['fixed'][v] = values[v][3]
        
        self.logger.debug('Fit parameters set to %s',self.fitpar)

    # ======================================================================= #
    def set_fitresult(self,values):
        """
            Set fit results. Values is output of fitting routine. It is a list 
            of tuples
            [(parname),(par),(err-),(err+),chi,fnpointer]
        """
        self.parnames = values[0]
        
        for i in range(len(self.parnames)):
            key = values[0][i]
            self.fitpar['res'][key] = values[1][i]
            self.fitpar['dres-'][key] = values[2][i]
            self.fitpar['dres+'][key] = values[3][i]
            self.fitpar['chi'][key] = values[4]
        self.chi = values[4]
        self.fitfn = values[5]
        
        self.logger.debug('Setting fit results to %s',self.fitpar)
    

# ========================================================================== #
class temperature_class(object):
    """
        Emulate storage container for camp variable smpl_read_%
    """
    
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std
    
